import os
import tempfile
import timeit
import zipfile
import titanarchive

zip_file_path = None

EXTRACTION_ITERATIONS = 10

def titanarchive_extract():
    with titanarchive.TitanArchive(zip_file_path) as ta:
        for i in range(0, EXTRACTION_ITERATIONS):
            for i in range(0, ta.GetArchiveItemCount()):
                ta.ExtractArchiveItemToBufferByIndex(i)

def zipfile_extract():
    with zipfile.ZipFile(zip_file_path, 'r') as z:
        for i in range(0, EXTRACTION_ITERATIONS):
            for name in z.namelist():
                z.read(name)

def create_file(path, size):
    CHUNK_SIZE = 1024 * 1024
    with open(path, 'wb') as f:
        while size > 0:
            if CHUNK_SIZE > size:
                bytes_to_write = size
            else:
                bytes_to_write = CHUNK_SIZE
            f.write(os.urandom(bytes_to_write))
            size -= bytes_to_write
    
def benchmark_file(size):
    global zip_file_path
    print('Benchmarking...')
    with tempfile.TemporaryDirectory() as tmp_dir:
        large_file_path = os.path.join(tmp_dir, 'file')
        create_file(large_file_path, size)
        with tempfile.TemporaryDirectory() as tmp_dir2:
            zip_file_path = os.path.join(tmp_dir2, 'file.zip')
            with zipfile.ZipFile(zip_file_path, 'w') as z:
                z.write(large_file_path)
            print('TitanArchive: ' + str(timeit.timeit('titanarchive_extract()', globals=globals(), number=1)) + 'sec')
            print('ZipFile: ' + str(timeit.timeit('zipfile_extract()', globals=globals(), number=1)) + 'sec')
            zip_file_path = None

def benchmark_many_files():
    pass

if __name__ == '__main__':
    benchmark_file(1024*1024*1024)


import setuptools
import os
import subprocess
import tempfile
import shutil
import platform
import json
import hashlib
import tarfile
import sysconfig
from urllib.request import urlretrieve
from wheel.bdist_wheel import bdist_wheel
from setuptools.command.sdist import sdist
from setuptools.command.build_clib import build_clib

VERSION = '2.1'

SUPPORTED_PLATFORMS = ['win-amd64', 'win32', 'linux-x86_64', 'linux-x86']

if os.name == 'nt':
    import vswhere

#####################
src_files = ['P7Zip.cpp', 'TitanArchive.cpp', 'Compat.cpp']
os_libs = []
if os.name == 'nt':
    os_libs += ['OleAut32.lib']
elif os.name == 'posix':
    os_libs += ['-ldl']
SETUP_DIR = os.path.dirname(os.path.relpath(__file__))
DEPS_DIR = os.path.relpath(os.path.join(SETUP_DIR, 'deps'))
SRC_DIR = os.path.relpath(os.path.join(SETUP_DIR, 'src'))
PACKAGE_DIR = os.path.join(SRC_DIR, 'titanarchive')
BIN_DIR = os.path.join(SETUP_DIR, 'build')
#####################

if 'microsoft' in platform.uname()[3].lower() and SETUP_DIR.startswith('/mnt/c'):
    raise Exception('setup.py should not be run within a windows mount because of permission issues')

src_files = [os.path.join(SRC_DIR, path) for path in src_files]
bin_files = []

try:
    os.mkdir(BIN_DIR)
except FileExistsError:
    pass

def windows_set_env(vs_path, arch):
    vs_path = os.path.join(vs_path, 'VC\\Auxiliary\\Build')
    cmd = [os.path.join(vs_path, 'vcvars{}.bat'.format(arch)), '&&', 'set']
    popen = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = popen.communicate()
    if popen.wait() != 0:
        raise Exception(stderr.decode('mbcs'))
    output = stdout.decode('mbcs').split('\r\n')
    os.environ.update(dict((e[0].upper(), e[1]) for e in [p.rstrip().split('=', 1) for p in output] if len(e) == 2))

def windows_build(arch, debug=False):
    vs_path = vswhere.find(latest=True, requires='Microsoft.VisualStudio.Component.VC.Tools.x86.x64', prop='installationPath')
    if len(vs_path) == 0:
        raise Exception('Unable to discover installation path of Visual Studio C/C++ build tools')

    vs_path = vs_path[0]
    msvc_path = os.path.join(vs_path, 'VC\\Tools\\MSVC\\', open(os.path.join(vs_path, 'VC\\Auxiliary\\Build\\Microsoft.VCToolsVersion.default.txt'),'r').read().rstrip())

    host_arch = 'x64' if 'PROGRAMFILES(X86)' in os.environ else 'x86'
    cl_path32 = os.path.join(msvc_path, 'bin\\Host{}\\x86\\cl.exe'.format(host_arch))
    cl_path64 = os.path.join(msvc_path, 'bin\\Host{}\\x64\\cl.exe'.format(host_arch))
    cl_args = ['/LD', '/EHsc', '/MT', '/MP', '/W4', '/WX']

    if debug:
        cl_args += ['/Od', '/DEBUG', '/Z7']
    else:
        cl_args.append('/O2')

    with tempfile.TemporaryDirectory() as tmpdirname:
        windows_set_env(vs_path, arch)

        extract_deps(tmpdirname)
        p7zip_build_dir = os.path.join(tmpdirname, '7z-src', 'CPP', '7zip', 'Bundles', 'Format7zF')
        subprocess.run(['nmake'], cwd=p7zip_build_dir, check=True)
        bin_file = os.path.join(BIN_DIR, 'TitanArchive{}-7z.dll'.format(arch))
        shutil.copy2(os.path.join(p7zip_build_dir, 'x86' if arch == 32 else 'x64', '7z.dll'), bin_file)
        bin_files.append(os.path.relpath(bin_file))

        cl_args.append('/Fo{}\\'.format(tmpdirname))
        bin_file = os.path.join(BIN_DIR, 'TitanArchive{}.dll'.format(arch))
        subprocess.run([eval('cl_path{}'.format(arch)), *cl_args, '/Fe:{}'.format(bin_file), *src_files, *os_libs], check=True)
        bin_files.append(os.path.relpath(bin_file))


def linux_build(arch, debug=False):
    gpp_path = shutil.which('g++')

    if gpp_path is None or shutil.which('cc') is None:
        raise Exception('Unable to build, missing cc and/or g++')

    with tempfile.TemporaryDirectory() as tmpdirname:
        extract_deps(tmpdirname)
        p7zip_build_dir = os.path.join(tmpdirname, '7z-src', 'CPP', '7zip', 'Bundles', 'Format7zF')

        if arch == 32:
            os.environ['USE_ASM'] = ''
            p7zip_build_args = ['make', '-f', '../../cmpl_gcc_x86.mak', '-e']
            out_dir = os.path.join(p7zip_build_dir, 'b', 'g_x86')
        elif arch == 64:
            p7zip_build_args = ['make', '-f', 'makefile.gcc']
            out_dir = os.path.join(p7zip_build_dir, '_o')
        else:
            raise Exception('Unknown architecture specified')

        subprocess.run(p7zip_build_args, cwd=p7zip_build_dir, check=True)
        bin_file = os.path.join(BIN_DIR, 'TitanArchive{}-7z.so'.format(arch))
        shutil.copy2(os.path.join(out_dir, '7z.so'), bin_file)
        bin_files.append(os.path.relpath(bin_file))

    gpp_args = ['-shared', '-fPIC', '-Wall', '-std=c++14']

    if debug:
        gpp_args += ['-O0', '-g3']
    else:
        gpp_args += ['-O3', '-g0']

    if arch == 32:
        gpp_args += ['-m32']
        bin_file = os.path.join(BIN_DIR, 'TitanArchive32.so')
    elif arch == 64:
        bin_file = os.path.join(BIN_DIR, 'TitanArchive64.so')

    subprocess.run([gpp_path, *gpp_args, *src_files, '-o', bin_file, *os_libs], check=True)
    bin_files.append(os.path.relpath(bin_file))

def sha256_file(path):
    h = hashlib.sha256()
    with open(path, 'rb') as f:
        for chunk in iter(lambda: f.read(4096), b''):
            h.update(chunk)
    return h.hexdigest()

def download_deps():
    for dep in json.load(open(os.path.join(DEPS_DIR, 'manifest.json'), 'r')):
        full_path = os.path.join(DEPS_DIR, dep['filename'])
        try:
            if dep['sha256'] == sha256_file(full_path):
                continue
        except:
            pass

        print('Downloading: {}'.format(dep['url']))
        urlretrieve(dep['url'], full_path)

def extract_deps(path):
    for dep in json.load(open(os.path.join(DEPS_DIR, 'manifest.json'), 'r')):
        extract_dir = os.path.join(path, dep['filename'].split('.', 1)[0])
        if dep['filename'].endswith('.tar.xz'):
            os.mkdir(extract_dir)
            with tarfile.open(os.path.join(DEPS_DIR, dep['filename'])) as f:
                f.extractall(extract_dir)
        else:
            raise Exception('Unsupported archive')

def build_code(plat_name, debug):
    if plat_name == 'win-amd64':
        windows_build(64, debug)
    elif plat_name == 'win32':
        windows_build(32, debug)
    elif plat_name == 'linux-x86_64':
        linux_build(64, debug)
    elif plat_name == 'linux-x86':
        linux_build(32, debug)
    else:
        raise Exception('Unknown / Missing build platform specified')

class TitanArchiveBuildCLib(build_clib):
    user_options = [('debug', 'g', 'compile with debugging information'), ('platform=', 'p', 'specify the target platform')]

    def run(self):
        download_deps()
        build_code(self.platform, False if self.debug is None else True)

    def initialize_options(self):
        build_clib.initialize_options(self)
        self.platform = None

class TitanArchiveBDistWheel(bdist_wheel):
    def run(self):
        if not self.plat_name_supplied:
            self.plat_name = sysconfig.get_platform()
            self.plat_name_supplied = True
        if self.plat_name not in SUPPORTED_PLATFORMS:
            raise Exception('Invalid platform name. Options: {}'.format(SUPPORTED_PLATFORMS))
        download_deps()
        build_code(self.plat_name, False)
        bdist_wheel.run(self)

class TitanArchiveSDist(sdist):
    def run(self):
        download_deps()
        sdist.run(self)

setuptools.setup(
    name = 'titanarchive',
    version = VERSION,
    description = 'An easy to use multiplatform archive extraction wrapper library for 7-Zip',
    long_description = open(os.path.join(SETUP_DIR, 'README.md'), 'r').read(),
    long_description_content_type = 'text/markdown',
    license = 'LGPL-2.1',
    url = 'https://github.com/4d61726b/TitanArchive',
    packages = setuptools.find_packages(where=SRC_DIR),
    package_dir = {'': SRC_DIR},
    python_requires = '>=3',
    cmdclass = {'bdist_wheel': TitanArchiveBDistWheel, 'sdist': TitanArchiveSDist, 'build_clib': TitanArchiveBuildCLib},
    data_files = [('DLLs' if os.name == 'nt' else 'lib', bin_files)],
)



import unittest
import titanarchive
import os
import tempfile
import zipfile
import time
import datetime
import math
from enum import Enum, auto

ARCHIVE_PATH = os.path.dirname(os.path.realpath(__file__))
TEST_ZIP = os.path.join(ARCHIVE_PATH, 'archives', 'Test.zip')
PW_TEST_ZIP = os.path.join(ARCHIVE_PATH, 'archives', 'Pw.zip')

if os.name == 'nt':
    SLASH_CHAR = '\\'
else:
    SLASH_CHAR = '/'

class ArchiveInputType(Enum):
    MEMORY = auto()
    DISK = auto()
    FD = auto()

def filetime_to_unix_epoch(ft):
    EPOCH_AS_FILETIME = 116444736000000000
    us = (ft - EPOCH_AS_FILETIME) // 10
    return math.ceil(datetime.timedelta(microseconds = us).total_seconds())

def open_with_invalid_format(cls, input_type):
    try:
        extract_and_verify(cls, input_type, TEST_ZIP, 'invalidformat')
        raise Exception('Unreachable')
    except titanarchive.TitanArchiveException:
        pass

def open_with_wrong_format(cls, input_type):
    try:
        extract_and_verify(cls, input_type, TEST_ZIP, 'rar')
        raise Exception('Unreachable')
    except titanarchive.TitanArchiveException:
        pass

def extract_and_verify(cls, input_type, zip_on_disk, archive_format = None, password = None):
    if input_type == ArchiveInputType.MEMORY:
        with open(zip_on_disk, 'rb') as f:
            data = f.read()
            ta = titanarchive.TitanArchive(data, password, archive_format)
            _extract_and_verify(cls, zip_on_disk, ta, password)
    elif input_type == ArchiveInputType.DISK:
        ta = titanarchive.TitanArchive(zip_on_disk, password, archive_format)
        _extract_and_verify(cls, zip_on_disk, ta, password)
    elif input_type == ArchiveInputType.FD:
        fd = os.open(zip_on_disk, os.O_RDONLY | (os.O_BINARY if os.name == 'nt' else 0))
        try:
            ta = titanarchive.TitanArchive(fd, password, archive_format)
        finally:
            os.close(fd)
        _extract_and_verify(cls, zip_on_disk, ta, password)
    else:
        raise Exception('Unknown type specified')

def _extract_and_verify(cls, zip_file, ta, password):
    with tempfile.TemporaryDirectory() as tmp_dir:
        with zipfile.ZipFile(zip_file, 'r') as z:
            if password is not None:
                z.setpassword(password.encode())
            z.extractall(tmp_dir)
            for f in z.infolist():
                name, date_time, = f.filename, f.date_time
                name = os.path.join(tmp_dir, os.path.normpath(name))
                date_time = time.mktime(date_time + (0, 0, -1))
                os.utime(name, (date_time, date_time))
            cls.assertEqual(compare(cls, tmp_dir, ta, ''), ta.GetArchiveItemCount())

def _compare_archive_items(cls, ai1, ai2):
    cls.assertEqual(ai1.Index, ai2.Index)
    if SLASH_CHAR in ai1.Path and SLASH_CHAR not in ai2.Path:
        cls.assertEqual(os.path.basename(ai1.Path), ai2.Path)
    elif SLASH_CHAR in ai2.Path and SLASH_CHAR not in ai1.Path:
        cls.assertEqual(ai1.Path, os.path.basename(ai2.Path))
    else:
        cls.assertEqual(ai1.Path, ai2.Path)
    cls.assertEqual(ai1.IsDir, ai2.IsDir)
    cls.assertEqual(ai1.Size, ai2.Size)
    cls.assertEqual(ai1.MTime, ai2.MTime)

def compare(cls, disk_dir, ta, curr_dir):
    cls.assertEqual(len(os.listdir(disk_dir)), len(ta.ListDirectory(curr_dir)))
    total_elements = 0
    for entry in os.listdir(disk_dir):
        full_entry = os.path.join(disk_dir, entry)
        found = False
        for entry2 in ta.ListDirectory(curr_dir):
            if entry == entry2.Path:
                cls.assertEqual(os.path.isdir(full_entry), entry2.IsDir)
                if not os.path.isdir(full_entry):
                    cls.assertEqual(os.path.getsize(full_entry), entry2.Size)
                cls.assertAlmostEqual(os.path.getmtime(full_entry), filetime_to_unix_epoch(entry2.MTime), delta=1)
                if os.path.isfile(full_entry):
                    with open(full_entry, 'rb') as f:
                        data1 = f.read()
                        data2 = ta.ExtractArchiveItemToBufferByPath(os.path.join(curr_dir, entry2.Path))
                        cls.assertEqual(data1, data2.getvalue())
                        data2 = ta.ExtractArchiveItemToBufferByIndex(entry2.Index)
                        cls.assertEqual(data1, data2.getvalue())
                        _compare_archive_items(cls, ta.GetArchiveItemPropertiesByIndex(entry2.Index), entry2)
                        _compare_archive_items(cls, ta.GetArchiveItemPropertiesByPath(os.path.join(curr_dir, entry2.Path)), entry2)
                found = True
                break
        cls.assertTrue(found)
        total_elements += 1
        if os.path.isdir(full_entry):
            total_elements += compare(cls, full_entry, ta, os.path.join(curr_dir, entry))
    return total_elements

class ExtractionTests(unittest.TestCase):
    @classmethod
    def setUp(self):
        titanarchive.GlobalInitialize()
    @classmethod
    def tearDown(self):
        titanarchive.GlobalUninitialize()

    def test_GlobalInitMulti(self):
        titanarchive.GlobalUninitialize()
        titanarchive.GlobalInitialize()
        titanarchive.GlobalInitialize()
    def test_GlobalUninitMulti(self):
        titanarchive.GlobalUninitialize()
        titanarchive.GlobalUninitialize()
    def test_GlobalInitPath(self):
        if os.name == 'nt':
            titanarchive.GlobalInitialize('C:\\Program Files\\7-Zip\\7z.dll')
        elif os.name == 'posix':
            titanarchive.GlobalInitialize('/usr/lib/p7zip/7z.so')
        else:
            raise Exception('Unknown Platform')
        extract_and_verify(self, ArchiveInputType.MEMORY, TEST_ZIP)
    def test_GlobalArchiveFormats(self):
        self.assertTrue(len(titanarchive.GlobalGetSupportedArchiveFormats()) > 50)
    def test_OpenFromMemory(self):
        extract_and_verify(self, ArchiveInputType.MEMORY, TEST_ZIP)
        extract_and_verify(self, ArchiveInputType.MEMORY, TEST_ZIP, 'zip')
        extract_and_verify(self, ArchiveInputType.MEMORY, PW_TEST_ZIP, password = 'password')
        extract_and_verify(self, ArchiveInputType.MEMORY, TEST_ZIP, password = 'password')
        open_with_invalid_format(self, ArchiveInputType.MEMORY)
        open_with_wrong_format(self, ArchiveInputType.MEMORY)
    def test_OpenFromDisk(self):
        extract_and_verify(self, ArchiveInputType.DISK, TEST_ZIP)
        extract_and_verify(self, ArchiveInputType.DISK, TEST_ZIP, 'zip')
        extract_and_verify(self, ArchiveInputType.DISK, PW_TEST_ZIP, password = 'password')
        extract_and_verify(self, ArchiveInputType.DISK, TEST_ZIP, password = 'password')
        open_with_invalid_format(self, ArchiveInputType.DISK)
        open_with_wrong_format(self, ArchiveInputType.DISK)
    def test_OpenFromFD(self):
        extract_and_verify(self, ArchiveInputType.FD, TEST_ZIP)
        extract_and_verify(self, ArchiveInputType.FD, TEST_ZIP, 'zip')
        extract_and_verify(self, ArchiveInputType.FD, PW_TEST_ZIP, password = 'password')
        extract_and_verify(self, ArchiveInputType.FD, TEST_ZIP, password = 'password')
        open_with_invalid_format(self, ArchiveInputType.FD)
        open_with_wrong_format(self, ArchiveInputType.FD)
    def test_GetArchiveFormat(self):
        with titanarchive.TitanArchive(TEST_ZIP) as ta:
            self.assertEqual(ta.GetArchiveFormat(), 'zip')
    def test_PasswordIterations(self):
        i = 0
        with titanarchive.TitanArchive(PW_TEST_ZIP) as ta:
            try:
                while True:
                    if not ta.GetArchiveItemPropertiesByIndex(i).IsDir:
                        break
                    i += 1
            except titanarchive.TitanArchiveException:
                raise Exception('No files discovered in archive')
        with titanarchive.TitanArchive(PW_TEST_ZIP) as ta:
            data = ta.ExtractArchiveItemToBufferByIndex(i, password = 'password')
            try:
                data = ta.ExtractArchiveItemToBufferByIndex(i, password = 'wrongpassword')
            except titanarchive.TitanArchiveException:
                pass
            else:
                raise Exception('Unreachable')
            try:
                data = ta.ExtractArchiveItemToBufferByIndex(i)
            except titanarchive.TitanArchiveException:
                pass
            else:
                raise Exception('Unreachable')
        with titanarchive.TitanArchive(PW_TEST_ZIP, password = 'wrongpassword') as ta:
            data = ta.ExtractArchiveItemToBufferByIndex(i, password = 'password')
            try:
                data = ta.ExtractArchiveItemToBufferByIndex(i)
            except titanarchive.TitanArchiveException:
                pass
            else:
                raise Exception('Unreachable')
            try:
                data = ta.ExtractArchiveItemToBufferByIndex(i, password = 'wrongpassword')
            except titanarchive.TitanArchiveException:
                pass
            else:
                raise Exception('Unreachable')
        with titanarchive.TitanArchive(PW_TEST_ZIP, password = 'password') as ta:
            data = ta.ExtractArchiveItemToBufferByIndex(i, password = 'password')
            data = ta.ExtractArchiveItemToBufferByIndex(i)
            try:
                data = ta.ExtractArchiveItemToBufferByIndex(i, password = 'wrongpassword')
            except titanarchive.TitanArchiveException:
                pass
            else:
                raise Exception('Unreachable')
    def test_MiscInvalid(self):
        with titanarchive.TitanArchive(TEST_ZIP) as ta:
            try:
                ta.ListDirectory('Invalid Directory')
                raise Exception('Unreachable')
            except titanarchive.TitanArchiveException:
                pass
            try:
                ta.GetArchiveItemPropertiesByPath('Invalid Path')
                raise Exception('Unreachable')
            except titanarchive.TitanArchiveException:
                pass
            try:
                ta.GetArchiveItemPropertiesByIndex(99999999)
                raise Exception('Unreachable')
            except titanarchive.TitanArchiveException:
                pass
            try:
                ta.ExtractArchiveItemToBufferByPath('Invalid Path')
                raise Exception('Unreachable')
            except titanarchive.TitanArchiveException:
                pass
            try:
                ta.ExtractArchiveItemToBufferByIndex(99999999)
                raise Exception('Unreachable')
            except titanarchive.TitanArchiveException:
                pass
    
    def test_CloseArchiveAndException(self):
        ta = titanarchive.TitanArchive(TEST_ZIP)    
        ta.ListDirectory('')
        ta.CloseArchive()
        try:
            ta.ListDirectory('')
            raise Exception('Unreachable')
        except titanarchive.TitanArchiveException:
            pass
                    
if __name__ == '__main__':
    unittest.main(verbosity=2)
    


import ctypes
from io import BytesIO
from collections import namedtuple
import os
import sys
if os.name == 'nt':
    from ctypes import wintypes
    import msvcrt

ARCHIVER_STATUS_SUCCESS = 0
ARCHIVER_STATUS_FAILURE = 1
E_FAIL = 0x80004005

if os.name == 'nt':
    ext = '.dll'
elif os.name == 'posix':
    ext = '.so'
else:
    raise Exception('Unknown platform')
if ctypes.sizeof(ctypes.c_voidp) == 4:
    bits = '32'
elif ctypes.sizeof(ctypes.c_voidp) == 8:
    bits = '64'
else:
    raise Exception('Unknown architecture')

TITAN_ARCHIVE_MODULE_NAME = 'TitanArchive{}{}'.format(bits, ext)
TITAN_ARCHIVE_7Z_MODULE_NAME = 'TitanArchive{}-7z{}'.format(bits, ext)

if sys.prefix == sys.base_prefix:
    if os.name == 'nt':
        TITAN_ARCHIVE_MODULE_DIR = os.path.join(sys.prefix, 'DLLs')
    else:
        TITAN_ARCHIVE_MODULE_DIR = os.path.join('/', 'usr', 'local', 'lib')
    TITAN_ARCHIVE_MODULE = os.path.join(TITAN_ARCHIVE_MODULE_DIR, TITAN_ARCHIVE_MODULE_NAME)
else:
    if os.name == 'nt':
        mod_dir = 'DLLs'
    else:
        mod_dir = 'lib'

    TITAN_ARCHIVE_MODULE_DIR = os.path.join(sys.prefix, mod_dir)
    TITAN_ARCHIVE_MODULE = os.path.join(TITAN_ARCHIVE_MODULE_DIR, TITAN_ARCHIVE_MODULE_NAME)
    if not os.path.isfile(TITAN_ARCHIVE_MODULE):
        TITAN_ARCHIVE_MODULE_DIR = os.path.join(sys.base_prefix, mod_dir)
        TITAN_ARCHIVE_MODULE = os.path.join(TITAN_ARCHIVE_MODULE_DIR, TITAN_ARCHIVE_MODULE_NAME)

if not os.path.isfile(TITAN_ARCHIVE_MODULE):
    raise Exception('Unable to find TitanArchive support module (%s)' % TITAN_ARCHIVE_MODULE_NAME)

TITAN_ARCHIVE_7Z_MODULE = os.path.join(TITAN_ARCHIVE_MODULE_DIR, TITAN_ARCHIVE_7Z_MODULE_NAME)

if not os.path.isfile(TITAN_ARCHIVE_7Z_MODULE):
    raise Exception('Unable to find TitanArchive 7z support module (%s)' % TITAN_ARCHIVE_7Z_MODULE)

class TitanArchiveException(Exception):
    def __init__(self, hr, err_string):
        self.hr = hr
        self.err_string = err_string
        super().__init__(self.err_string)

class _ArchiveItem(ctypes.Structure):
    _fields_ = [('Index', ctypes.c_uint),
                ('Path', ctypes.c_wchar_p),
                ('IsDir', ctypes.c_bool),
                ('Size', ctypes.c_ulonglong),
                ('MTime', ctypes.c_ulonglong),
                ('_ItemNext', ctypes.c_void_p)]

def _GetDict(st):
    d = dict((field, getattr(st, field)) for field, _ in st._fields_ if not field.startswith('_'))
    ArchiveItem = namedtuple('ArchiveItem', d)
    return ArchiveItem(**d)

def _GetNamedTuple(struct):
    rtn = collections.namedtuple()

class TitanArchive():

    def __del__(self):
        self._DeleteArchiveContext()

    def __init__(self, archive, password = None, archive_format = None):
        self.archive = archive
        self.password = password
        self.archive_format = archive_format
        self._ctx = ctypes.c_void_p(lib.CreateArchiveContext())
        
        if self._ctx.value == ctypes.c_void_p(0).value:
            raise TitanArchiveException(*GetGlobalError())
        if isinstance(self.archive, int):
            self.OpenArchiveFD(self.archive, self.password, self.archive_format)
        elif isinstance(self.archive, str):
            self.OpenArchiveDisk(self.archive, self.password, self.archive_format)
        else:
            self.OpenArchiveMemory(self.archive, self.password, self.archive_format)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self._DeleteArchiveContext()

    def __iter__(self):
        for i in range(0, self.GetArchiveItemCount()):
            yield self.GetArchiveItemPropertiesByIndex(i)

    def OpenArchiveMemory(self, buf, password = None, archive_format = None):
        size = len(buf)
        if lib.OpenArchiveMemory(self._ctx, buf, ctypes.c_ulonglong(size), ctypes.c_wchar_p(password), ctypes.c_wchar_p(archive_format)) != ARCHIVER_STATUS_SUCCESS:
            raise TitanArchiveException(*self.GetError())

    def OpenArchiveDisk(self, path, password = None, archive_format = None):
        if lib.OpenArchiveDisk(self._ctx, ctypes.c_wchar_p(path), ctypes.c_wchar_p(password), ctypes.c_wchar_p(archive_format)) != ARCHIVER_STATUS_SUCCESS:
            raise TitanArchiveException(*self.GetError())

    def OpenArchiveFD(self, fd, password = None, archive_format = None):
        if os.name == 'nt':
            if lib.OpenArchiveHandle(self._ctx, wintypes.HANDLE(msvcrt.get_osfhandle(fd)), ctypes.c_wchar_p(password), ctypes.c_wchar_p(archive_format)) != ARCHIVER_STATUS_SUCCESS:
                raise TitanArchiveException(*self.GetError())
        else:
            if lib.OpenArchiveFD(self._ctx, ctypes.c_int(fd), ctypes.c_wchar_p(password), ctypes.c_wchar_p(archive_format)) != ARCHIVER_STATUS_SUCCESS:
                raise TitanArchiveException(*self.GetError())

    def GetArchiveFormat(self):
        rtn = ctypes.c_wchar_p()
        if lib.GetArchiveFormat(self._ctx, ctypes.byref(rtn)) != ARCHIVER_STATUS_SUCCESS:
            raise TitanArchiveException(*self.GetError())
        return rtn.value

    def GetArchiveItemCount(self):
        rtn = ctypes.c_uint()
        if lib.GetArchiveItemCount(self._ctx, ctypes.byref(rtn)) != ARCHIVER_STATUS_SUCCESS:
            raise TitanArchiveException(*self.GetError())
        return rtn.value

    def ListDirectory(self, path = ''):
        rtn = []
        ai = ctypes.POINTER(_ArchiveItem)()
        count = ctypes.c_ulonglong()
        if lib.ListDirectory(self._ctx, ctypes.c_wchar_p(path), ctypes.byref(ai), ctypes.byref(count)) != ARCHIVER_STATUS_SUCCESS:
            raise TitanArchiveException(*self.GetError())
        cur = ai
        for i in range(0, count.value):
            cur = cur.contents
            rtn.append(_GetDict(cur))
            cur = ctypes.cast(cur._ItemNext, ctypes.POINTER(_ArchiveItem))
        self._FreeArchiveItem(ai)
        return rtn

    def GetArchiveItemPropertiesByPath(self, path):
        ai = ctypes.POINTER(_ArchiveItem)()
        if lib.GetArchiveItemPropertiesByPath(self._ctx, ctypes.c_wchar_p(path), ctypes.byref(ai)) != ARCHIVER_STATUS_SUCCESS:
            raise TitanArchiveException(*self.GetError())
        ai = ai.contents
        rtn = _GetDict(ai)
        self._FreeArchiveItem(ai)
        return rtn

    def GetArchiveItemPropertiesByIndex(self, index):
        ai = ctypes.POINTER(_ArchiveItem)()
        if lib.GetArchiveItemPropertiesByIndex(self._ctx, ctypes.c_uint(index), ctypes.byref(ai)) != ARCHIVER_STATUS_SUCCESS:
            raise TitanArchiveException(*self.GetError())
        ai = ai.contents
        rtn = _GetDict(ai)
        self._FreeArchiveItem(ai)
        return rtn

    def ExtractArchiveItemToBufferByIndex(self, index, password = None):
        size = self.GetArchiveItemPropertiesByIndex(index).Size
        rtn = BytesIO(bytearray(size))
        if size and lib.ExtractArchiveItemToBufferByIndex(self._ctx, ctypes.c_uint(index), rtn.getvalue(), ctypes.c_ulonglong(size), ctypes.c_wchar_p(password)) != ARCHIVER_STATUS_SUCCESS:
            raise TitanArchiveException(*self.GetError())
        return rtn

    def ExtractArchiveItemToBufferByPath(self, path, password = None):
        size = self.GetArchiveItemPropertiesByPath(path).Size
        rtn = BytesIO(bytearray(size))
        if size and lib.ExtractArchiveItemToBufferByPath(self._ctx, ctypes.c_wchar_p(path), ctypes.c_char_p(rtn.getvalue()), ctypes.c_ulonglong(size), ctypes.c_wchar_p(password)) != ARCHIVER_STATUS_SUCCESS:
            raise TitanArchiveException(*self.GetError())
        return rtn

    def CloseArchive(self):
        if lib.CloseArchive(self._ctx) != ARCHIVER_STATUS_SUCCESS:
            raise TitanArchiveException(*self.GetError())

    def _FreeArchiveItem(self, ai):
        if lib.FreeArchiveItem(self._ctx, ai) != ARCHIVER_STATUS_SUCCESS:
            raise TitanArchiveException(*self.GetError())

    def _DeleteArchiveContext(self):
        if self._ctx.value != ctypes.c_void_p(0).value:
            if lib.DeleteArchiveContext(self._ctx) != ARCHIVER_STATUS_SUCCESS:
                raise TitanArchiveException(*self.GetError())
            else:
                self._ctx = ctypes.c_void_p(0)

    def GetError(self):
        hr = ctypes.c_int()
        err = ctypes.c_wchar_p()
        if lib.GetError(self._ctx, ctypes.byref(hr), ctypes.byref(err)) != ARCHIVER_STATUS_SUCCESS:
            raise TitanArchiveException(E_FAIL, 'Unable to get error')
        return (hr.value, err.value)

def GetGlobalError():
    hr = ctypes.c_int()
    err = ctypes.c_wchar_p()
    if lib.GetError(None, ctypes.byref(hr), ctypes.byref(err)) != ARCHIVER_STATUS_SUCCESS:
        raise TitanArchiveException(E_FAIL, 'Unable to get error')
    return (hr.value, err.value)

def GlobalInitialize(lib_path = TITAN_ARCHIVE_7Z_MODULE):
    if lib.GlobalInitialize(lib_path) != ARCHIVER_STATUS_SUCCESS:
        raise TitanArchiveException(*GetGlobalError())

def GlobalAddCodec(format, lib_path):
    if lib.GlobalAddCodec(format, lib_path) != ARCHIVER_STATUS_SUCCESS:
        raise TitanArchiveException(*GetGlobalError())

def GlobalGetSupportedArchiveFormats():
    af = ctypes.c_wchar_p()
    if lib.GlobalGetSupportedArchiveFormats(ctypes.byref(af)) != ARCHIVER_STATUS_SUCCESS:
        raise TitanArchiveException(*GetGlobalError())
    return af.value.split(',') if len(af.value) > 0 else []

def GlobalUninitialize():
    lib.GlobalUninitialize()

lib = ctypes.CDLL(TITAN_ARCHIVE_MODULE)

# ARCHIVER_STATUS GlobalInitialize(const wchar_t* wszLibPath)
lib.GlobalInitialize.argtypes = [ctypes.c_wchar_p]
lib.GlobalInitialize.restype = ctypes.c_uint

# ARCHIVER_STATUS GlobalAddCodec(const wchar_t* wszFormat, const wchar_t* wszLibPath);
lib.GlobalAddCodec.argtype = [ctypes.c_wchar_p, ctypes.c_wchar_p]
lib.GlobalAddCodec.restype = ctypes.c_uint

# ARCHIVER_STATUS GlobalGetSupportedArchiveFormats(const wchar_t** pwszFormats);
lib.GlobalGetSupportedArchiveFormats.argtype = [ctypes.POINTER(ctypes.c_wchar_p)]
lib.GlobalGetSupportedArchiveFormats.restype = ctypes.c_uint

# void GlobalUninitialize()
lib.GlobalUninitialize.argtypes = []

# void* CreateArchiveContext()
lib.CreateArchiveContext.restype = ctypes.c_void_p

# ARCHIVER_STATUS OpenArchiveMemory(void* pCtx, uint8_t* pBuf, uint64_t ui64BufSize, const wchar_t* wszPassword, const wchar_t* wszFormat)
lib.OpenArchiveMemory.argtypes = [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_ulonglong, ctypes.c_wchar_p, ctypes.c_wchar_p]
lib.OpenArchiveMemory.restype = ctypes.c_uint

# ARCHIVER_STATUS OpenArchiveDisk(void* pCtx, const wchar_t* wszPath, const wchar_t* wszPassword, const wchar_t* wszFormat)
lib.OpenArchiveDisk.argtypes = [ctypes.c_void_p, ctypes.c_wchar_p, ctypes.c_wchar_p, ctypes.c_wchar_p]
lib.OpenArchiveDisk.restype = ctypes.c_uint

if os.name == 'nt':
    # ARCHIVER_STATUS OpenArchiveHandle(void* pCtx, HANDLE hArchive, const wchar_t* wszPassword, const wchar_t* wszFormat)
    lib.OpenArchiveHandle.argtypes = [ctypes.c_void_p, wintypes.HANDLE, ctypes.c_wchar_p, ctypes.c_wchar_p]
    lib.OpenArchiveHandle.restype = ctypes.c_uint
else:
    # ARCHIVER_STATUS OpenArchiveFD(void* pCtx, int iFd, const wchar_t* wszPassword, const wchar_t* wszFormat)
    lib.OpenArchiveFD.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_wchar_p, ctypes.c_wchar_p]
    lib.OpenArchiveFD.restype = ctypes.c_uint

# ARCHIVER_STATUS GetArchiveFormat(void* pCtx, const wchar_t** pwszFormat)
lib.GetArchiveFormat.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_wchar_p)]
lib.GetArchiveFormat.restype = ctypes.c_uint

# ARCHIVER_STATUS GetArchiveItemCount(void* pCtx, uint32_t* pArchiveItemCount)
lib.GetArchiveItemCount.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_uint)]
lib.GetArchiveItemCount.restype = ctypes.c_uint

# ARCHIVER_STATUS ListDirectory(void* pCtx, const wchar_t* wszPath, ArchiveItem** ppItems, uint64_t* pItemCount)
lib.ListDirectory.argtypes = [ctypes.c_void_p, ctypes.c_wchar_p, ctypes.POINTER(ctypes.POINTER(_ArchiveItem)), ctypes.POINTER(ctypes.c_ulonglong)]
lib.ListDirectory.restype = ctypes.c_uint

# ARCHIVER_STATUS GetArchiveItemPropertiesByPath(void* pCtx, const wchar_t* wszPath, ArchiveItem** ppItem)
lib.GetArchiveItemPropertiesByPath.argtypes = [ctypes.c_void_p, ctypes.c_wchar_p, ctypes.POINTER(ctypes.POINTER(_ArchiveItem))]
lib.GetArchiveItemPropertiesByPath.restype = ctypes.c_uint

# ARCHIVER_STATUS GetArchiveItemPropertiesByIndex(void* pCtx, uint32_t ui32ItemIndex, ArchiveItem** ppItem)
lib.GetArchiveItemPropertiesByIndex.argtypes = [ctypes.c_void_p, ctypes.c_uint, ctypes.POINTER(ctypes.POINTER(_ArchiveItem))]
lib.GetArchiveItemPropertiesByIndex.restype = ctypes.c_uint

# ARCHIVER_STATUS FreeArchiveItem(void* pCtx, ArchiveItem* pItem)
lib.FreeArchiveItem.argtypes = [ctypes.c_void_p, ctypes.POINTER(_ArchiveItem)]
lib.FreeArchiveItem.restype = ctypes.c_uint

# ARCHIVER_STATUS ExtractArchiveItemToBufferByIndex(void* pCtx, uint32_t ui32ItemIndex, uint8_t* pBuf, uint64_t ui64BufSize, const wchar_t* wszPassword)
lib.ExtractArchiveItemToBufferByIndex.argtypes = [ctypes.c_void_p, ctypes.c_uint, ctypes.c_void_p, ctypes.c_ulonglong, ctypes.c_wchar_p]
lib.ExtractArchiveItemToBufferByIndex.restype = ctypes.c_uint

# ARCHIVER_STATUS ExtractArchiveItemToBufferByPath(void* pCtx, const wchar_t* wszPath, uint8_t* pBuf, uint64_t ui64BufSize, const wchar_t* wszPassword)
lib.ExtractArchiveItemToBufferByPath.argtypes = [ctypes.c_void_p, ctypes.c_wchar_p, ctypes.c_void_p, ctypes.c_ulonglong, ctypes.c_wchar_p]
lib.ExtractArchiveItemToBufferByPath.restype = ctypes.c_uint

# ARCHIVER_STATUS CloseArchive(void* pCtx)
lib.CloseArchive.argtypes = [ctypes.c_void_p]
lib.CloseArchive.restype = ctypes.c_uint

# ARCHIVER_STATUS DeleteArchiveContext(void* pCtx)
lib.DeleteArchiveContext.argtypes = [ctypes.c_void_p]
lib.DeleteArchiveContext.restype = ctypes.c_uint

# ARCHIVER_STATUS GetError(void* pCtx, HRESULT* pHr, const wchar_t** ppError)
lib.GetError.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_wchar_p)]
lib.GetError.restype = ctypes.c_uint

GlobalInitialize()
